@@ -234,7 +234,7 @@ struct error_entry ceer_errors[] = {
 	{ 127,	"Interworking, unspecified" },
 };
 
-gboolean valid_phone_number_format(const char *number)
+gboolean valid_number_format(const char *number, int length)
 {
 	int len = strlen(number);
 	int begin = 0;
@@ -246,7 +246,7 @@ gboolean valid_phone_number_format(const char *number)
 	if (number[0] == '+')
 		begin = 1;
 
-	if ((len - begin) > OFONO_MAX_PHONE_NUMBER_LENGTH)
+	if ((len - begin) > length)
 		return FALSE;
 
 	for (i = begin; i < len; i++) {
@@ -262,6 +262,46 @@ gboolean valid_phone_number_format(const char *number)
 	return TRUE;
 }
 
+/*
+ * According to 3GPP TS 24.011 or 3GPP TS 31.102, some
+ * addresses (or numbers), like Service Centre address,
+ * Destination address, or EFADN (Abbreviated dialling numbers),
+ * are up 20 digits.
+ */
+gboolean valid_phone_number_format(const char *number)
+{
+	return valid_number_format(number, 20);
+}
+
+gboolean valid_long_phone_number_format(const char *number)
+{
+	return valid_number_format(number, OFONO_MAX_PHONE_NUMBER_LENGTH);
+}
+
+gboolean valid_cdma_phone_number_format(const char *number)
+{
+	int len = strlen(number);
+	int i;
+
+	if (!len)
+		return FALSE;
+
+	if (len > OFONO_CDMA_MAX_PHONE_NUMBER_LENGTH)
+		return FALSE;
+
+	for (i = 0; i < len; i++) {
+		if (number[i] >= '0' && number[i] <= '9')
+			continue;
+
+		if (number[i] == '*' || number[i] == '#')
+			continue;
+
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
 const char *telephony_error_to_str(const struct ofono_error *error)
 {
 	struct error_entry *e;
@@ -405,6 +445,23 @@ void string_to_phone_number(const char *str, struct ofono_phone_number *ph)
 	}
 }
 
+const char *cdma_phone_number_to_string(
+				const struct ofono_cdma_phone_number *ph)
+{
+	static char buffer[OFONO_CDMA_MAX_PHONE_NUMBER_LENGTH + 1];
+
+	strncpy(buffer, ph->number, OFONO_CDMA_MAX_PHONE_NUMBER_LENGTH);
+	buffer[OFONO_CDMA_MAX_PHONE_NUMBER_LENGTH] = '\0';
+
+	return buffer;
+}
+
+void string_to_cdma_phone_number(const char *str,
+					struct ofono_cdma_phone_number *ph)
+{
+	strcpy(ph->number, str);
+}
+
 gboolean valid_ussd_string(const char *str, gboolean call_in_progress)
 {
 	int len = strlen(str);
@@ -516,7 +573,7 @@ gboolean parse_ss_control_string(char *str, int *ss_type,
 	/* Must have at least one other '#' */
 	c = strrchr(str+cur, '#');
 
-	if (!c)
+	if (c == NULL)
 		goto out;
 
 	*dn = c+1;
@@ -592,43 +649,6 @@ const char *bearer_class_to_string(enum bearer_class cls)
 	return NULL;
 }
 
-gboolean is_valid_pin(const char *pin, enum pin_type type)
-{
-	unsigned int i;
-
-	/* Pin must not be empty */
-	if (pin == NULL || pin[0] == '\0')
-		return FALSE;
-
-	i = strlen(pin);
-	if (i != strspn(pin, "0123456789"))
-		return FALSE;
-
-	switch (type) {
-	case PIN_TYPE_PIN:
-		/* 11.11 Section 9.3 ("CHV"): 4..8 IA-5 digits */
-		if (4 <= i && i <= 8)
-			return TRUE;
-		break;
-	case PIN_TYPE_PUK:
-		/* 11.11 Section 9.3 ("UNBLOCK CHV"), 8 IA-5 digits */
-		if (i == 8)
-			return TRUE;
-		break;
-	case PIN_TYPE_NET:
-		/* 22.004 Section 5.2, 4 IA-5 digits */
-		if (i == 4)
-			return TRUE;
-		break;
-	case PIN_TYPE_NONE:
-		if (i < 8)
-			return TRUE;
-		break;
-	}
-
-	return FALSE;
-}
-
 const char *registration_status_to_string(int status)
 {
 	switch (status) {
@@ -705,3 +725,10 @@ const char *ofono_uuid_to_str(const struct ofono_uuid *uuid)
 
 	return encode_hex_own_buf(uuid->uuid, OFONO_SHA1_UUID_LEN, 0, buf);
 }
+
+void ofono_call_init(struct ofono_call *call)
+{
+	memset(call, 0, sizeof(struct ofono_call));
+	call->cnap_validity = CNAP_VALIDITY_NOT_AVAILABLE;
+	call->clip_validity = CLIP_VALIDITY_NOT_AVAILABLE;
+}
