@@ -45,12 +45,12 @@
 #define SIM_CACHE_BASEPATH STORAGEDIR "/%s-%i"
 #define SIM_CACHE_VERSION SIM_CACHE_BASEPATH "/version"
 #define SIM_CACHE_PATH SIM_CACHE_BASEPATH "/%04x"
-#define SIM_CACHE_HEADER_SIZE 38
-#define SIM_FILE_INFO_SIZE 6
+#define SIM_CACHE_HEADER_SIZE 39
+#define SIM_FILE_INFO_SIZE 7
 #define SIM_IMAGE_CACHE_BASEPATH STORAGEDIR "/%s-%i/images"
 #define SIM_IMAGE_CACHE_PATH SIM_IMAGE_CACHE_BASEPATH "/%d.xpm"
 
-#define SIM_FS_VERSION 1
+#define SIM_FS_VERSION 2
 
 static gboolean sim_fs_op_next(gpointer user_data);
 static gboolean sim_fs_op_read_record(gpointer user);
@@ -69,6 +69,7 @@ struct sim_fs_op {
 	gconstpointer cb;
 	gboolean is_read;
 	void *userdata;
+	struct ofono_sim_context *context;
 };
 
 static void sim_fs_op_free(struct sim_fs_op *node)
@@ -84,10 +85,14 @@ struct sim_fs {
 	int fd;
 	struct ofono_sim *sim;
 	const struct ofono_sim_driver *driver;
+	GSList *contexts;
 };
 
 void sim_fs_free(struct sim_fs *fs)
 {
+	if (fs == NULL)
+		return;
+
 	if (fs->op_source) {
 		g_source_remove(fs->op_source);
 		fs->op_source = 0;
@@ -102,9 +107,29 @@ void sim_fs_free(struct sim_fs *fs)
 		g_queue_free(fs->op_q);
 	}
 
+	if (fs->contexts != NULL) {
+		GSList *l;
+
+		for (l = fs->contexts; l; l = l->next) {
+			struct ofono_sim_context *context = l->data;
+
+			sim_fs_context_free(context);
+		}
+	}
+
 	g_free(fs);
 }
 
+struct file_watch {
+	struct ofono_watchlist_item item;
+	int ef;
+};
+
+struct ofono_sim_context {
+	struct sim_fs *fs;
+	struct ofono_watchlist *file_watches;
+};
+
 struct sim_fs *sim_fs_new(struct ofono_sim *sim,
 				const struct ofono_sim_driver *driver)
 {
@@ -121,6 +146,99 @@ struct sim_fs *sim_fs_new(struct ofono_sim *sim,
 	return fs;
 }
 
+struct ofono_sim_context *sim_fs_context_new(struct sim_fs *fs)
+{
+	struct ofono_sim_context *context =
+		g_try_new0(struct ofono_sim_context, 1);
+
+	if (context == NULL)
+		return NULL;
+
+	context->fs = fs;
+	fs->contexts = g_slist_prepend(fs->contexts, context);
+
+	return context;
+}
+
+void sim_fs_context_free(struct ofono_sim_context *context)
+{
+	struct sim_fs *fs = context->fs;
+	int n = 0;
+	struct sim_fs_op *op;
+
+	while ((op = g_queue_peek_nth(fs->op_q, n)) != NULL) {
+		if (op->context != context) {
+			n += 1;
+			continue;
+		}
+
+		if (n == 0) {
+			op->cb = NULL;
+
+			n += 1;
+			continue;
+		}
+
+		sim_fs_op_free(op);
+		g_queue_remove(fs->op_q, op);
+	}
+
+	if (context->file_watches)
+		__ofono_watchlist_free(context->file_watches);
+
+	fs->contexts = g_slist_remove(fs->contexts, context);
+	g_free(context);
+}
+
+unsigned int sim_fs_file_watch_add(struct ofono_sim_context *context, int id,
+					ofono_sim_file_changed_cb_t cb,
+					void *userdata,
+					ofono_destroy_func destroy)
+{
+	struct file_watch *watch;
+
+	if (cb == NULL)
+		return 0;
+
+	if (context->file_watches == NULL)
+		context->file_watches = __ofono_watchlist_new(g_free);
+
+	watch = g_new0(struct file_watch, 1);
+
+	watch->ef = id;
+	watch->item.notify = cb;
+	watch->item.notify_data = userdata;
+	watch->item.destroy = destroy;
+
+	return __ofono_watchlist_add_item(context->file_watches,
+					(struct ofono_watchlist_item *) watch);
+}
+
+void sim_fs_file_watch_remove(struct ofono_sim_context *context,
+				unsigned int id)
+{
+	__ofono_watchlist_remove_item(context->file_watches, id);
+}
+
+void sim_fs_notify_file_watches(struct sim_fs *fs, int id)
+{
+	GSList *l;
+
+	for (l = fs->contexts; l; l = l->next) {
+		struct ofono_sim_context *context = l->data;
+		GSList *k;
+
+		for (k = context->file_watches->items; k; k = k->next) {
+			struct file_watch *w = k->data;
+			ofono_sim_file_changed_cb_t notify = w->item.notify;
+
+			if (id == -1 || w->ef == id)
+				notify(w->ef, w->item.notify_data);
+		}
+	}
+
+}
+
 static void sim_fs_end_current(struct sim_fs *fs)
 {
 	struct sim_fs_op *op = g_queue_pop_head(fs->op_q);
@@ -142,6 +260,11 @@ static void sim_fs_op_error(struct sim_fs *fs)
 {
 	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
 
+	if (op->cb == NULL) {
+		sim_fs_end_current(fs);
+		return;
+	}
+
 	if (op->info_only == TRUE)
 		((sim_fs_read_info_cb_t) op->cb)
 			(0, 0, 0, 0, op->userdata);
@@ -201,6 +324,11 @@ static void sim_fs_op_write_cb(const struct ofono_error *error, void *data)
 	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
 	ofono_sim_file_write_cb_t cb = op->cb;
 
+	if (cb == NULL) {
+		sim_fs_end_current(fs);
+		return;
+	}
+
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR)
 		cb(1, op->userdata);
 	else
@@ -247,6 +375,11 @@ static void sim_fs_op_read_block_cb(const struct ofono_error *error,
 	memcpy(op->buffer + bufoff, data + dataoff, tocopy);
 	cache_block(fs, op->current, 256, data, len);
 
+	if (op->cb == NULL) {
+		sim_fs_end_current(fs);
+		return;
+	}
+
 	op->current++;
 
 	if (op->current > end_block) {
@@ -269,6 +402,13 @@ static gboolean sim_fs_op_read_block(gpointer user_data)
 	int end_block;
 	unsigned short read_bytes;
 
+	fs->op_source = 0;
+
+	if (op->cb == NULL) {
+		sim_fs_end_current(fs);
+		return FALSE;
+	}
+
 	start_block = op->offset / 256;
 	end_block = (op->offset + (op->num_bytes - 1)) / 256;
 
@@ -355,11 +495,16 @@ static void sim_fs_op_retrieve_cb(const struct ofono_error *error,
 		return;
 	}
 
-	cb(1, op->length, op->current, data, op->record_length, op->userdata);
-
 	cache_block(fs, op->current - 1, op->record_length,
 			data, op->record_length);
 
+	if (cb == NULL) {
+		sim_fs_end_current(fs);
+		return;
+	}
+
+	cb(1, op->length, op->current, data, op->record_length, op->userdata);
+
 	if (op->current < total) {
 		op->current += 1;
 		fs->op_source = g_idle_add(sim_fs_op_read_record, fs);
@@ -378,6 +523,11 @@ static gboolean sim_fs_op_read_record(gpointer user)
 
 	fs->op_source = 0;
 
+	if (op->cb == NULL) {
+		sim_fs_end_current(fs);
+		return FALSE;
+	}
+
 	while (fs->fd != -1 && op->current <= total) {
 		int offset = (op->current - 1) / 8;
 		int bit = 1 << ((op->current - 1) % 8);
@@ -434,14 +584,14 @@ static gboolean sim_fs_op_read_record(gpointer user)
 	return FALSE;
 }
 
-static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
-				enum ofono_sim_file_structure structure,
-				int record_length,
-				const unsigned char access[3],
-				unsigned char file_status,
-				void *data)
+static void sim_fs_op_cache_fileinfo(struct sim_fs *fs,
+					const struct ofono_error *error,
+					int length,
+					enum ofono_sim_file_structure structure,
+					int record_length,
+					const unsigned char access[3],
+					unsigned char file_status)
 {
-	struct sim_fs *fs = data;
 	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
 	const char *imsi = ofono_sim_get_imsi(fs->sim);
 	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
@@ -452,26 +602,11 @@ static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
 	gboolean cache;
 	char *path;
 
-	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		sim_fs_op_error(fs);
-		return;
-	}
-
-	if (structure != op->structure) {
-		ofono_error("Requested file structure differs from SIM: %x",
-				op->id);
-		sim_fs_op_error(fs);
-		return;
-	}
-
 	/* TS 11.11, Section 9.3 */
 	update = file_access_condition_decode(access[0] & 0xf);
 	rehabilitate = file_access_condition_decode((access[2] >> 4) & 0xf);
 	invalidate = file_access_condition_decode(access[2] & 0xf);
 
-	op->structure = structure;
-	op->length = length;
-
 	/* Never cache card holder writable files */
 	cache = (update == SIM_FILE_ACCESS_ADM ||
 			update == SIM_FILE_ACCESS_NEVER) &&
@@ -480,6 +615,67 @@ static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
 			(rehabilitate == SIM_FILE_ACCESS_ADM ||
 				rehabilitate == SIM_FILE_ACCESS_NEVER);
 
+	if (imsi == NULL || phase == OFONO_SIM_PHASE_UNKNOWN || cache == FALSE)
+		return;
+
+	memset(fileinfo, 0, SIM_CACHE_HEADER_SIZE);
+
+	fileinfo[0] = error->type;
+	fileinfo[1] = length >> 8;
+	fileinfo[2] = length & 0xff;
+	fileinfo[3] = structure;
+	fileinfo[4] = record_length >> 8;
+	fileinfo[5] = record_length & 0xff;
+	fileinfo[6] = file_status;
+
+	path = g_strdup_printf(SIM_CACHE_PATH, imsi, phase, op->id);
+	fs->fd = TFR(open(path, O_WRONLY | O_CREAT | O_TRUNC, SIM_CACHE_MODE));
+	g_free(path);
+
+	if (fs->fd == -1)
+		return;
+
+	if (TFR(write(fs->fd, fileinfo, SIM_CACHE_HEADER_SIZE)) ==
+			SIM_CACHE_HEADER_SIZE)
+		return;
+
+	TFR(close(fs->fd));
+	fs->fd = -1;
+}
+
+static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
+				enum ofono_sim_file_structure structure,
+				int record_length,
+				const unsigned char access[3],
+				unsigned char file_status,
+				void *data)
+{
+	struct sim_fs *fs = data;
+	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		sim_fs_op_error(fs);
+		return;
+	}
+
+	sim_fs_op_cache_fileinfo(fs, error, length, structure, record_length,
+					access, file_status);
+
+	if (structure != op->structure) {
+		ofono_error("Requested file structure differs from SIM: %x",
+				op->id);
+		sim_fs_op_error(fs);
+		return;
+	}
+
+	if (op->cb == NULL) {
+		sim_fs_end_current(fs);
+		return;
+	}
+
+	op->structure = structure;
+	op->length = length;
+
 	if (structure == OFONO_SIM_FILE_STRUCTURE_TRANSPARENT) {
 		if (op->num_bytes == 0)
 			op->num_bytes = op->length;
@@ -499,8 +695,8 @@ static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
 
 	if (op->info_only == TRUE) {
 		/*
-		 * It's info-only request. So there is no need to request
-		 * actual contents of the EF-files. Just return the EF-info.
+		 * It's an info-only request, so there is no need to request
+		 * actual contents of the EF. Just return the EF-info.
 		 */
 		sim_fs_read_info_cb_t cb = op->cb;
 
@@ -508,35 +704,7 @@ static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
 			op->record_length, op->userdata);
 
 		sim_fs_end_current(fs);
-
-		return;
 	}
-
-	if (imsi == NULL || phase == OFONO_SIM_PHASE_UNKNOWN || cache == FALSE)
-		return;
-
-	memset(fileinfo, 0, SIM_CACHE_HEADER_SIZE);
-
-	fileinfo[0] = error->type;
-	fileinfo[1] = length >> 8;
-	fileinfo[2] = length & 0xff;
-	fileinfo[3] = structure;
-	fileinfo[4] = record_length >> 8;
-	fileinfo[5] = record_length & 0xff;
-
-	path = g_strdup_printf(SIM_CACHE_PATH, imsi, phase, op->id);
-	fs->fd = TFR(open(path, O_RDWR | O_CREAT | O_TRUNC, SIM_CACHE_MODE));
-	g_free(path);
-
-	if (fs->fd == -1)
-		return;
-
-	if (TFR(write(fs->fd, fileinfo, SIM_CACHE_HEADER_SIZE)) ==
-			SIM_CACHE_HEADER_SIZE)
-		return;
-
-	TFR(close(fs->fd));
-	fs->fd = -1;
 }
 
 static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
@@ -544,7 +712,6 @@ static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
 	const char *imsi = ofono_sim_get_imsi(fs->sim);
 	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
 	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
-	gboolean ret = FALSE;
 	char *path;
 	int fd;
 	ssize_t len;
@@ -553,9 +720,9 @@ static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
 	int file_length;
 	enum ofono_sim_file_structure structure;
 	int record_length;
+	unsigned char file_status;
 
-	if (imsi == NULL || phase == OFONO_SIM_PHASE_UNKNOWN ||
-			op->info_only == TRUE)
+	if (imsi == NULL || phase == OFONO_SIM_PHASE_UNKNOWN)
 		return FALSE;
 
 	path = g_strdup_printf(SIM_CACHE_PATH, imsi, phase, op->id);
@@ -584,6 +751,7 @@ static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
 	file_length = (fileinfo[1] << 8) | fileinfo[2];
 	structure = fileinfo[3];
 	record_length = (fileinfo[4] << 8) | fileinfo[5];
+	file_status = fileinfo[6];
 
 	if (structure == OFONO_SIM_FILE_STRUCTURE_TRANSPARENT)
 		record_length = file_length;
@@ -598,10 +766,23 @@ static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
 	fs->fd = fd;
 
 	if (error_type != OFONO_ERROR_TYPE_NO_ERROR ||
-			structure != op->structure)
+			structure != op->structure) {
 		sim_fs_op_error(fs);
+		return TRUE;
+	}
 
-	if (structure == OFONO_SIM_FILE_STRUCTURE_TRANSPARENT) {
+	if (op->info_only == TRUE) {
+		/*
+		 * It's an info-only request, so there is no need to request
+		 * actual contents of the EF. Just return the EF-info.
+		 */
+		sim_fs_read_info_cb_t cb = op->cb;
+
+		cb(1, file_status, op->length,
+			op->record_length, op->userdata);
+
+		sim_fs_end_current(fs);
+	} else if (structure == OFONO_SIM_FILE_STRUCTURE_TRANSPARENT) {
 		if (op->num_bytes == 0)
 			op->num_bytes = op->length;
 
@@ -616,7 +797,7 @@ static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
 
 error:
 	TFR(close(fd));
-	return ret;
+	return FALSE;
 }
 
 static gboolean sim_fs_op_next(gpointer user_data)
@@ -632,6 +813,11 @@ static gboolean sim_fs_op_next(gpointer user_data)
 
 	op = g_queue_peek_head(fs->op_q);
 
+	if (op->cb == NULL) {
+		sim_fs_end_current(fs);
+		return FALSE;
+	}
+
 	if (op->is_read == TRUE) {
 		if (sim_fs_op_check_cached(fs))
 			return FALSE;
@@ -666,10 +852,11 @@ static gboolean sim_fs_op_next(gpointer user_data)
 	return FALSE;
 }
 
-int sim_fs_read_info(struct sim_fs *fs, int id,
+int sim_fs_read_info(struct ofono_sim_context *context, int id,
 			enum ofono_sim_file_structure expected_type,
 			sim_fs_read_info_cb_t cb, void *data)
 {
+	struct sim_fs *fs = context->fs;
 	struct sim_fs_op *op;
 
 	if (cb == NULL)
@@ -694,6 +881,7 @@ int sim_fs_read_info(struct sim_fs *fs, int id,
 	op->userdata = data;
 	op->is_read = TRUE;
 	op->info_only = TRUE;
+	op->context = context;
 
 	g_queue_push_tail(fs->op_q, op);
 
@@ -703,11 +891,12 @@ int sim_fs_read_info(struct sim_fs *fs, int id,
 	return 0;
 }
 
-int sim_fs_read(struct sim_fs *fs, int id,
+int sim_fs_read(struct ofono_sim_context *context, int id,
 		enum ofono_sim_file_structure expected_type,
 		unsigned short offset, unsigned short num_bytes,
 		ofono_sim_file_read_cb_t cb, void *data)
 {
+	struct sim_fs *fs = context->fs;
 	struct sim_fs_op *op;
 
 	if (cb == NULL)
@@ -734,6 +923,7 @@ int sim_fs_read(struct sim_fs *fs, int id,
 	op->offset = offset;
 	op->num_bytes = num_bytes;
 	op->info_only = FALSE;
+	op->context = context;
 
 	g_queue_push_tail(fs->op_q, op);
 
@@ -743,10 +933,12 @@ int sim_fs_read(struct sim_fs *fs, int id,
 	return 0;
 }
 
-int sim_fs_write(struct sim_fs *fs, int id, ofono_sim_file_write_cb_t cb,
+int sim_fs_write(struct ofono_sim_context *context, int id,
+			ofono_sim_file_write_cb_t cb,
 			enum ofono_sim_file_structure structure, int record,
 			const unsigned char *data, int length, void *userdata)
 {
+	struct sim_fs *fs = context->fs;
 	struct sim_fs_op *op;
 	gconstpointer fn = NULL;
 
@@ -788,6 +980,7 @@ int sim_fs_write(struct sim_fs *fs, int id, ofono_sim_file_write_cb_t cb,
 	op->structure = structure;
 	op->length = length;
 	op->current = record;
+	op->context = context;
 
 	g_queue_push_tail(fs->op_q, op);
 
@@ -907,9 +1100,6 @@ void sim_fs_check_version(struct sim_fs *fs)
 	const char *imsi = ofono_sim_get_imsi(fs->sim);
 	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
 	unsigned char version;
-	struct dirent **entries;
-	int len;
-	char *path;
 
 	if (imsi == NULL || phase == OFONO_SIM_PHASE_UNKNOWN)
 		return;
@@ -918,10 +1108,20 @@ void sim_fs_check_version(struct sim_fs *fs)
 		if (version == SIM_FS_VERSION)
 			return;
 
-	path = g_strdup_printf(SIM_CACHE_BASEPATH, imsi, phase);
+	sim_fs_cache_flush(fs);
+
+	version = SIM_FS_VERSION;
+	write_file(&version, 1, SIM_CACHE_MODE, SIM_CACHE_VERSION, imsi, phase);
+}
+
+void sim_fs_cache_flush(struct sim_fs *fs)
+{
+	const char *imsi = ofono_sim_get_imsi(fs->sim);
+	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
+	char *path = g_strdup_printf(SIM_CACHE_BASEPATH, imsi, phase);
+	struct dirent **entries;
+	int len = scandir(path, &entries, NULL, alphasort);
 
-	ofono_info("Detected old simfs version in %s, removing", path);
-	len = scandir(path, &entries, NULL, alphasort);
 	g_free(path);
 
 	if (len > 0) {
@@ -934,20 +1134,47 @@ void sim_fs_check_version(struct sim_fs *fs)
 		g_free(entries);
 	}
 
-	path = g_strdup_printf(SIM_IMAGE_CACHE_BASEPATH, imsi, phase);
-	len = scandir(path, &entries, NULL, alphasort);
+	sim_fs_image_cache_flush(fs);
+}
+
+void sim_fs_cache_flush_file(struct sim_fs *fs, int id)
+{
+	const char *imsi = ofono_sim_get_imsi(fs->sim);
+	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
+	char *path = g_strdup_printf(SIM_CACHE_PATH, imsi, phase, id);
+
+	remove(path);
+	g_free(path);
+}
+
+void sim_fs_image_cache_flush(struct sim_fs *fs)
+{
+	const char *imsi = ofono_sim_get_imsi(fs->sim);
+	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
+	char *path = g_strdup_printf(SIM_IMAGE_CACHE_BASEPATH, imsi, phase);
+	struct dirent **entries;
+	int len = scandir(path, &entries, NULL, alphasort);
+
 	g_free(path);
 
-	if (len > 0) {
-		/* Remove everything */
-		while (len--) {
-			remove_imagefile(imsi, phase, entries[len]);
-			g_free(entries[len]);
-		}
+	if (len <= 0)
+		return;
 
-		g_free(entries);
+	/* Remove everything */
+	while (len--) {
+		remove_imagefile(imsi, phase, entries[len]);
+		g_free(entries[len]);
 	}
 
-	version = SIM_FS_VERSION;
-	write_file(&version, 1, SIM_CACHE_MODE, SIM_CACHE_VERSION, imsi, phase);
+	g_free(entries);
+}
+
+void sim_fs_image_cache_flush_file(struct sim_fs *fs, int id)
+{
+	const char *imsi = ofono_sim_get_imsi(fs->sim);
+	enum ofono_sim_phase phase = ofono_sim_get_phase(fs->sim);
+	char *path = g_strdup_printf(SIM_IMAGE_CACHE_PATH, imsi, phase, id);
+
+	remove(path);
+	g_free(path);
 }
