@@ -99,6 +99,11 @@ struct extern_req {
 static void envelope_queue_run(struct ofono_stk *stk);
 static void timers_update(struct ofono_stk *stk);
 
+#define ADD_ERROR_RESULT(result, error, addn_info)		\
+		result.type = error;				\
+		result.additional_len = sizeof(addn_info);	\
+		result.additional = addn_info;			\
+
 static int stk_respond(struct ofono_stk *stk, struct stk_response *rsp,
 			ofono_stk_generic_cb_t cb)
 {
@@ -474,7 +479,12 @@ static void stk_alpha_id_set(struct ofono_stk *stk,
 {
 	char *alpha = dbus_apply_text_attributes(text, attr);
 
-	if (alpha || icon)
+	/*
+	 * Currently, we are treating null data object(len = 0, no value part)
+	 * and no alpha identifier cases equally. This may be changed once
+	 * better idea is found out.
+	 */
+	if (alpha != NULL)
 		stk_agent_display_action_info(stk->current_agent, alpha, icon);
 
 	g_free(alpha);
@@ -857,7 +867,10 @@ static gboolean handle_command_send_sms(const struct stk_command *cmd,
 	msg_list.next = NULL;
 
 	if (__ofono_sms_txq_submit(sms, &msg_list, 0, &uuid, NULL, NULL) < 0) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		unsigned char no_cause_result[] = { 0x00 };
+
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -1093,26 +1106,26 @@ static gboolean handle_command_set_up_menu(const struct stk_command *cmd,
 						struct stk_response *rsp,
 						struct ofono_stk *stk)
 {
-	gboolean modified = FALSE;
-
-	if (stk->main_menu) {
-		stk_menu_free(stk->main_menu);
-		stk->main_menu = NULL;
-
-		modified = TRUE;
-	}
+	struct stk_menu *menu = NULL;
 
 	if (cmd->setup_menu.items) {
-		stk->main_menu = stk_menu_create_from_set_up_menu(cmd);
+		menu = stk_menu_create_from_set_up_menu(cmd);
 
-		if (stk->main_menu)
-			modified = TRUE;
-		else
+		if (menu == NULL) {
 			rsp->result.type = STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD;
+			return TRUE;
+		}
 	}
 
-	if (modified)
-		emit_menu_changed(stk);
+	if (menu == NULL && stk->main_menu == NULL)
+		return TRUE;
+
+	if (stk->main_menu)
+		stk_menu_free(stk->main_menu);
+
+	stk->main_menu = menu;
+
+	emit_menu_changed(stk);
 
 	return TRUE;
 }
@@ -1135,8 +1148,9 @@ static void request_selection_cb(enum stk_agent_result result, uint8_t id,
 	switch (result) {
 	case STK_AGENT_RESULT_OK:
 	{
-		static struct ofono_error error =
-			{ .type = OFONO_ERROR_TYPE_FAILURE };
+		static struct ofono_error error = {
+			.type = OFONO_ERROR_TYPE_FAILURE
+		};
 		struct stk_response rsp;
 
 		memset(&rsp, 0, sizeof(rsp));
@@ -1183,9 +1197,12 @@ static gboolean handle_command_select_item(const struct stk_command *cmd,
 					request_selection_cb, stk,
 					request_selection_destroy,
 					stk->timeout * 1000) < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
 		request_selection_destroy(stk);
 
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -1258,9 +1275,8 @@ static void display_text_cb(enum stk_agent_result result, void *user_data)
 
 	case STK_AGENT_RESULT_BUSY:
 		memset(&rsp, 0, sizeof(rsp));
-		rsp.result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp.result.additional_len = sizeof(screen_busy_result);
-		rsp.result.additional = screen_busy_result;
+		ADD_ERROR_RESULT(rsp.result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					screen_busy_result);
 		if (stk_respond(stk, &rsp, stk_command_cb))
 			stk_command_cb(&error, stk);
 		break;
@@ -1301,7 +1317,10 @@ static gboolean handle_command_display_text(const struct stk_command *cmd,
 
 	/* We most likely got an out of memory error, tell SIM to retry */
 	if (err < 0) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		unsigned char no_cause_result[] = { 0x00 };
+
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -1475,11 +1494,14 @@ static gboolean handle_command_get_inkey(const struct stk_command *cmd,
 	g_free(text);
 
 	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
 		/*
 		 * We most likely got an out of memory error, tell SIM
 		 * to retry
 		 */
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -1564,11 +1586,14 @@ static gboolean handle_command_get_input(const struct stk_command *cmd,
 	g_free(text);
 
 	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
 		/*
 		 * We most likely got an out of memory error, tell SIM
 		 * to retry
 		 */
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -1588,9 +1613,9 @@ static void call_setup_connected(struct ofono_call *call, void *data)
 	if (call == NULL || call->status == CALL_STATUS_DISCONNECTED) {
 		memset(&rsp, 0, sizeof(rsp));
 
-		rsp.result.type = STK_RESULT_TYPE_NETWORK_UNAVAILABLE;
-		rsp.result.additional_len = sizeof(facility_rejected_result);
-		rsp.result.additional = facility_rejected_result;
+		ADD_ERROR_RESULT(rsp.result,
+					STK_RESULT_TYPE_NETWORK_UNAVAILABLE,
+					facility_rejected_result);
 
 		if (stk_respond(stk, &rsp, stk_command_cb))
 			stk_command_cb(&error, stk);
@@ -1689,9 +1714,8 @@ static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
 	if (err == -EBUSY) {
 		memset(&rsp, 0, sizeof(rsp));
 
-		rsp.result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp.result.additional_len = sizeof(busy_on_call_result);
-		rsp.result.additional = busy_on_call_result;
+		ADD_ERROR_RESULT(rsp.result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					busy_on_call_result);
 
 		if (stk_respond(stk, &rsp, stk_command_cb))
 			stk_command_cb(&error, stk);
@@ -1707,9 +1731,8 @@ static void confirm_call_cb(enum stk_agent_result result, gboolean confirm,
 
 	memset(&rsp, 0, sizeof(rsp));
 
-	rsp.result.type = STK_RESULT_TYPE_NETWORK_UNAVAILABLE;
-	rsp.result.additional_len = sizeof(no_cause_result);
-	rsp.result.additional = no_cause_result;
+	ADD_ERROR_RESULT(rsp.result, STK_RESULT_TYPE_NETWORK_UNAVAILABLE,
+				no_cause_result);
 
 	if (stk_respond(stk, &rsp, stk_command_cb))
 		stk_command_cb(&error, stk);
@@ -1752,9 +1775,8 @@ static gboolean handle_command_set_up_call(const struct stk_command *cmd,
 	}
 
 	if (__ofono_voicecall_is_busy(vc, qualifier >> 1)) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp->result.additional_len = sizeof(busy_on_call_result);
-		rsp->result.additional = busy_on_call_result;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					busy_on_call_result);
 		return TRUE;
 	}
 
@@ -1772,11 +1794,14 @@ static gboolean handle_command_set_up_call(const struct stk_command *cmd,
 	g_free(alpha_id);
 
 	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
 		/*
 		 * We most likely got an out of memory error, tell SIM
 		 * to retry
 		 */
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -1849,9 +1874,8 @@ static void send_ussd_callback(int error, int dcs, const unsigned char *msg,
 		break;
 
 	default:
-		rsp.result.type = STK_RESULT_TYPE_USSD_RETURN_ERROR;
-		rsp.result.additional_len = sizeof(no_cause);
-		rsp.result.additional = no_cause;
+		ADD_ERROR_RESULT(rsp.result, STK_RESULT_TYPE_USSD_RETURN_ERROR,
+					no_cause);
 
 		if (stk_respond(stk, &rsp, stk_command_cb))
 			stk_command_cb(&failure, stk);
@@ -1876,9 +1900,9 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 		struct ofono_call_forwarding *cf = __ofono_atom_get_data(atom);
 
 		if (__ofono_call_forwarding_is_busy(cf)) {
-			rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-			rsp->result.additional_len = sizeof(busy_on_ss_result);
-			rsp->result.additional = busy_on_ss_result;
+			ADD_ERROR_RESULT(rsp->result,
+						STK_RESULT_TYPE_TERMINAL_BUSY,
+						busy_on_ss_result);
 			return TRUE;
 		}
 	}
@@ -1888,9 +1912,9 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 		struct ofono_call_barring *cb = __ofono_atom_get_data(atom);
 
 		if (__ofono_call_barring_is_busy(cb)) {
-			rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-			rsp->result.additional_len = sizeof(busy_on_ss_result);
-			rsp->result.additional = busy_on_ss_result;
+			ADD_ERROR_RESULT(rsp->result,
+						STK_RESULT_TYPE_TERMINAL_BUSY,
+						busy_on_ss_result);
 			return TRUE;
 		}
 	}
@@ -1900,9 +1924,9 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 		struct ofono_call_settings *cs = __ofono_atom_get_data(atom);
 
 		if (__ofono_call_settings_is_busy(cs)) {
-			rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-			rsp->result.additional_len = sizeof(busy_on_ss_result);
-			rsp->result.additional = busy_on_ss_result;
+			ADD_ERROR_RESULT(rsp->result,
+						STK_RESULT_TYPE_TERMINAL_BUSY,
+						busy_on_ss_result);
 			return TRUE;
 		}
 	}
@@ -1915,9 +1939,8 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 
 	ussd = __ofono_atom_get_data(atom);
 	if (__ofono_ussd_is_busy(ussd)) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp->result.additional_len = sizeof(busy_on_ussd_result);
-		rsp->result.additional = busy_on_ussd_result;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					busy_on_ussd_result);
 		return TRUE;
 	}
 
@@ -1938,9 +1961,8 @@ static gboolean handle_command_send_ussd(const struct stk_command *cmd,
 	}
 
 	if (err == -EBUSY) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp->result.additional_len = sizeof(busy_on_ussd_result);
-		rsp->result.additional = busy_on_ussd_result;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					busy_on_ussd_result);
 		return TRUE;
 	}
 
@@ -2112,14 +2134,14 @@ static void dtmf_sent_cb(int error, void *user_data)
 	if (error == ENOENT) {
 		struct stk_response rsp;
 		static unsigned char not_in_speech_call_result[] = { 0x07 };
-		static struct ofono_error failure =
-			{ .type = OFONO_ERROR_TYPE_FAILURE };
+		static struct ofono_error failure = {
+			.type = OFONO_ERROR_TYPE_FAILURE
+		};
 
 		memset(&rsp, 0, sizeof(rsp));
 
-		rsp.result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp.result.additional_len = sizeof(not_in_speech_call_result);
-		rsp.result.additional = not_in_speech_call_result;
+		ADD_ERROR_RESULT(rsp.result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					not_in_speech_call_result);
 
 		if (stk_respond(stk, &rsp, stk_command_cb))
 			stk_command_cb(&failure, stk);
@@ -2170,20 +2192,14 @@ static gboolean handle_command_send_dtmf(const struct stk_command *cmd,
 
 	err = __ofono_voicecall_tone_send(vc, dtmf, dtmf_sent_cb, stk);
 
-	if (err == -EBUSY) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		return TRUE;
-	}
-
 	if (err == -ENOSYS) {
 		rsp->result.type = STK_RESULT_TYPE_NOT_CAPABLE;
 		return TRUE;
 	}
 
 	if (err == -ENOENT) {
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
-		rsp->result.additional_len = sizeof(not_in_speech_call_result);
-		rsp->result.additional = not_in_speech_call_result;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					not_in_speech_call_result);
 		return TRUE;
 	}
 
@@ -2193,11 +2209,14 @@ static gboolean handle_command_send_dtmf(const struct stk_command *cmd,
 	}
 
 	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
 		/*
 		 * We most likely got an out of memory error, tell SIM
 		 * to retry
 		 */
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -2329,11 +2348,88 @@ static gboolean handle_command_play_tone(const struct stk_command *cmd,
 	g_free(text);
 
 	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
 		/*
 		 * We most likely got an out of memory error, tell SIM
 		 * to retry
 		 */
-		rsp->result.type = STK_RESULT_TYPE_TERMINAL_BUSY;
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
+		return TRUE;
+	}
+
+	stk->respond_on_exit = TRUE;
+	stk->cancel_cmd = stk_request_cancel;
+
+	return FALSE;
+}
+
+static void confirm_launch_browser_cb(enum stk_agent_result result,
+					gboolean confirm,
+					void *user_data)
+{
+	struct ofono_stk *stk = user_data;
+	unsigned char no_cause[] = { 0x00 };
+	struct ofono_error failure = { .type = OFONO_ERROR_TYPE_FAILURE };
+	struct stk_response rsp;
+
+	stk->respond_on_exit = FALSE;
+
+	switch (result) {
+	case STK_AGENT_RESULT_TIMEOUT:
+		confirm = FALSE;
+		/* Fall through */
+
+	case STK_AGENT_RESULT_OK:
+		if (confirm)
+			break;
+		/* Fall through */
+
+	default:
+		memset(&rsp, 0, sizeof(rsp));
+		ADD_ERROR_RESULT(rsp.result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause);
+
+		if (stk_respond(stk, &rsp, stk_command_cb))
+			stk_command_cb(&failure, stk);
+
+		return;
+	}
+
+	send_simple_response(stk, STK_RESULT_TYPE_SUCCESS);
+}
+
+static gboolean handle_command_launch_browser(const struct stk_command *cmd,
+						struct stk_response *rsp,
+						struct ofono_stk *stk)
+{
+	const struct stk_command_launch_browser *lb = &cmd->launch_browser;
+	char *alpha_id;
+	int err;
+
+	alpha_id = dbus_apply_text_attributes(lb->alpha_id ? lb->alpha_id : "",
+							&lb->text_attr);
+	if (alpha_id == NULL) {
+		rsp->result.type = STK_RESULT_TYPE_DATA_NOT_UNDERSTOOD;
+		return TRUE;
+	}
+
+	err = stk_agent_confirm_launch_browser(stk->current_agent, alpha_id,
+						lb->icon_id.id, lb->url,
+						confirm_launch_browser_cb,
+						stk, NULL, stk->timeout * 1000);
+	g_free(alpha_id);
+
+	if (err < 0) {
+		unsigned char no_cause_result[] = { 0x00 };
+
+		/*
+		 * We most likely got an out of memory error, tell SIM
+		 * to retry
+		 */
+		ADD_ERROR_RESULT(rsp->result, STK_RESULT_TYPE_TERMINAL_BUSY,
+					no_cause_result);
 		return TRUE;
 	}
 
@@ -2530,6 +2626,11 @@ void ofono_stk_proactive_command_notify(struct ofono_stk *stk,
 							&rsp, stk);
 		break;
 
+	case STK_COMMAND_TYPE_LAUNCH_BROWSER:
+		respond = handle_command_launch_browser(stk->pending_cmd,
+							&rsp, stk);
+		break;
+
 	default:
 		rsp.result.type = STK_RESULT_TYPE_COMMAND_NOT_UNDERSTOOD;
 		break;
@@ -2745,7 +2846,7 @@ void ofono_stk_register(struct ofono_stk *stk)
 
 	__ofono_atom_register(stk->atom, stk_unregister);
 
-	stk->timeout = 600; /* 10 minutes */
+	stk->timeout = 180; /* 3 minutes */
 	stk->short_timeout = 20; /* 20 seconds */
 	stk->envelope_q = g_queue_new();
 }
