@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -38,6 +38,10 @@ const unsigned char invalid_gsm_extended_len[] = {
 	0x1b, 0x28, 0x1b
 };
 
+const unsigned char invalid_ucs2[] = {
+	0x03, 0x93, 0x00, 0x00
+};
+
 unsigned short gsm_to_unicode_map[] =
 {
 0x00,	0x0040,
@@ -337,11 +341,12 @@ unsigned short gsm_turkish_to_unicode_map[] =
 	((c) < 0x80 ? 1 : \
 	 ((c) < 0x800 ? 2 : 3))
 
-static void test_invalid()
+static void test_invalid(void)
 {
 	long nwritten;
 	long nread;
 	char *res;
+	unsigned char *gsm;
 
 	res = convert_gsm_to_utf8(invalid_gsm_extended, 0, &nread, &nwritten,
 					0);
@@ -362,9 +367,22 @@ static void test_invalid()
 					&nread, &nwritten, 0);
 	g_assert(res == NULL);
 	g_assert(nread == 3);
+
+	gsm = convert_ucs2_to_gsm(invalid_ucs2,
+					sizeof(invalid_ucs2),
+					&nread, &nwritten, 0);
+	g_assert(gsm == NULL);
+	g_assert(nread == 2);
+
+	nread = 0;
+	gsm = convert_ucs2_to_gsm(invalid_ucs2,
+					sizeof(invalid_ucs2) - 1,
+					&nread, &nwritten, 0);
+	g_assert(gsm == NULL);
+	g_assert(nread == 0);
 }
 
-static void test_valid()
+static void test_valid(void)
 {
 	long nwritten;
 	long nread;
@@ -412,6 +430,7 @@ static void test_valid()
 		g_assert(back);
 
 		g_assert(nwritten == size);
+
 		if (c & 0x1b00) {
 			g_assert(back[0] == 0x1b);
 			g_assert(back[1] == (c & 0x7f));
@@ -425,7 +444,7 @@ static void test_valid()
 	}
 }
 
-static void test_valid_turkish()
+static void test_valid_turkish(void)
 {
 	long nwritten;
 	long nread;
@@ -452,7 +471,8 @@ static void test_valid_turkish()
 			buf[0] = c & 0x7f;
 		}
 
-		res = convert_gsm_to_utf8_with_lang(buf, size, &nread, &nwritten, 0, 1, 1);
+		res = convert_gsm_to_utf8_with_lang(buf, size, &nread,
+							&nwritten, 0, 1, 1);
 		g_assert(res);
 
 		if (g_test_verbose())
@@ -468,11 +488,13 @@ static void test_valid_turkish()
 
 		g_assert(nwritten == UTF8_LENGTH(verify[0]));
 
-		back = convert_utf8_to_gsm_with_lang(res, -1, &nread, &nwritten, 0, 1, 1);
+		back = convert_utf8_to_gsm_with_lang(res, -1, &nread,
+							&nwritten, 0, 1, 1);
 
 		g_assert(back);
 
 		g_assert(nwritten == size);
+
 		if (c & 0x1b00) {
 			g_assert(back[0] == 0x1b);
 			g_assert(back[1] == (c & 0x7f));
@@ -489,10 +511,11 @@ static void test_valid_turkish()
 static const char hex_packed[] = "493A283D0795C3F33C88FE06C9CB6132885EC6D34"
 					"1EDF27C1E3E97E7207B3A0C0A5241E377BB1D"
 					"7693E72E";
-static const char expected[] = "It is easy to read text messages via AT commands.";
+static const char expected[] = "It is easy to read text messages via AT "
+				"commands.";
 static int reported_text_size = 49;
 
-static void test_decode_encode()
+static void test_decode_encode(void)
 {
 	const char *sms = hex_packed;
 	unsigned char *decoded, *packed;
@@ -591,7 +614,7 @@ static void test_decode_encode()
 	g_free(hex_packed);
 }
 
-static void test_pack_size()
+static void test_pack_size(void)
 {
 	unsigned char c1[] = { 'a' };
 	unsigned char c2[] = { 'a', 'b' };
@@ -653,12 +676,14 @@ static void test_pack_size()
 	g_free(packed);
 }
 
-static void test_cr_handling()
+static void test_cr_handling(void)
 {
 	unsigned char c7[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g' };
-	unsigned char c7_expected[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', '\r' };
+	unsigned char c7_expected[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g',
+					'\r' };
 	unsigned char c8[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', '\r' };
-	unsigned char c8_expected[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', '\r', '\r' };
+	unsigned char c8_expected[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g',
+					'\r', '\r' };
 
 	unsigned char *packed;
 	unsigned char *unpacked;
@@ -714,7 +739,7 @@ static void test_cr_handling()
 	g_free(packed);
 }
 
-static void test_sms_handling()
+static void test_sms_handling(void)
 {
 	unsigned char c7[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g' };
 
@@ -755,7 +780,7 @@ static void test_sms_handling()
 	g_free(packed);
 }
 
-static void test_offset_handling()
+static void test_offset_handling(void)
 {
 	unsigned char c7[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g' };
 	unsigned char c8[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h' };
@@ -855,7 +880,7 @@ static unsigned char sim_82_1[] = { 0x82, 0x05, 0x04, 0x00, 0x2D, 0xB3, 0xB4,
 static unsigned char sim_82_2[] = { 0x82, 0x05, 0xD8, 0x00, 0x2D, 0xB3, 0xB4,
 					0x2D, 0x31 };
 
-static void test_sim()
+static void test_sim(void)
 {
 	char *utf8;
 
@@ -867,17 +892,17 @@ static void test_sim()
 
 	utf8 = sim_string_to_utf8(sim_80_1, sizeof(sim_80_1));
 	g_assert(utf8);
-	g_assert(strcmp(utf8, "ono") == 0 );
+	g_assert(strcmp(utf8, "ono") == 0);
 	g_free(utf8);
 
 	utf8 = sim_string_to_utf8(sim_80_2, sizeof(sim_80_2));
 	g_assert(utf8);
-	g_assert(strcmp(utf8, "ono") == 0 );
+	g_assert(strcmp(utf8, "ono") == 0);
 	g_free(utf8);
 
 	utf8 = sim_string_to_utf8(sim_80_3, sizeof(sim_80_3));
 	g_assert(utf8);
-	g_assert(strcmp(utf8, "ono") == 0 );
+	g_assert(strcmp(utf8, "ono") == 0);
 	g_free(utf8);
 
 	utf8 = sim_string_to_utf8(sim_81_0, sizeof(sim_81_0));
@@ -890,7 +915,7 @@ static void test_sim()
 
 	utf8 = sim_string_to_utf8(sim_81_1, sizeof(sim_81_1));
 	g_assert(utf8);
-	g_assert(strcmp(utf8, "ono") == 0 );
+	g_assert(strcmp(utf8, "ono") == 0);
 	g_free(utf8);
 
 	utf8 = sim_string_to_utf8(sim_82_0, sizeof(sim_82_0));
@@ -905,19 +930,83 @@ static void test_sim()
 	g_assert(utf8 == NULL);
 }
 
+static void test_unicode_to_gsm(void)
+{
+	long nwritten;
+	long nread;
+	int i;
+	unsigned char *res;
+	char *utf8;
+	unsigned char buf[2];
+	unsigned char *back;
+	gunichar2 verify;
+
+	static int map_size =
+		sizeof(gsm_to_unicode_map) / sizeof(unsigned short) / 2;
+
+	for (i = 0; i < map_size; i++) {
+		unsigned short c = gsm_to_unicode_map[i*2+1];
+
+		buf[0] = c >> 8;
+		buf[1] = c & 0xff;
+
+		res = convert_ucs2_to_gsm(buf, 2, &nread, &nwritten, 0);
+		g_assert(res);
+
+		if (g_test_verbose())
+			g_print("nread:%ld, nwritten:%ld, %s\n",
+				nread, nwritten, res);
+
+		if (res[0] == 0x1B)
+			g_assert(nwritten == 2);
+		else
+			g_assert(nwritten == 1);
+
+		utf8 = g_convert((const gchar *) buf, 2,
+				"UTF-8", "UCS-2BE",
+				NULL, NULL, NULL);
+		g_assert(utf8);
+
+		back = convert_utf8_to_gsm(utf8, strlen(utf8), &nread,
+						&nwritten, 0);
+		g_assert(back);
+
+		if (back[0] == 0x1B) {
+			g_assert(nwritten == 2);
+			verify = back[0] << 8 | back[1];
+		} else {
+			g_assert(nwritten == 1);
+			verify = back[0];
+		}
+
+		if (g_test_verbose())
+			g_print("nwritten:%ld, verify: 0x%x\n",
+				nwritten, verify);
+
+		g_assert(verify == gsm_to_unicode_map[i*2]);
+
+		g_free(res);
+		g_free(back);
+		g_free(utf8);
+	}
+}
+
 int main(int argc, char **argv)
 {
 	g_test_init(&argc, &argv, NULL);
 
 	g_test_add_func("/testutil/Invalid Conversions", test_invalid);
 	g_test_add_func("/testutil/Valid Conversions", test_valid);
-	g_test_add_func("/testutil/Valid Turkish National Variant Conversions", test_valid_turkish);
+	g_test_add_func("/testutil/Valid Turkish National Variant Conversions",
+			test_valid_turkish);
 	g_test_add_func("/testutil/Decode Encode", test_decode_encode);
 	g_test_add_func("/testutil/Pack Size", test_pack_size);
 	g_test_add_func("/testutil/CBS CR Handling", test_cr_handling);
 	g_test_add_func("/testutil/SMS Handling", test_sms_handling);
 	g_test_add_func("/testutil/Offset Handling", test_offset_handling);
 	g_test_add_func("/testutil/SIM conversions", test_sim);
+	g_test_add_func("/testutil/Valid Unicode to GSM Conversion",
+			test_unicode_to_gsm);
 
 	return g_test_run();
 }
